import polars as pl

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class CsvReader(PolarsStep):
    """Local CSV file reader"""

    def __init__(self, path: str, has_headers: bool, lazy: bool = True):
        self.lazy = lazy
        self.headers = has_headers
        self.path = path

    async def process(self, ctx: SpinePolarsContext, df: PolarsDataFrame = None) -> PolarsDataFrame:

        ctx.logger.info(f"Reading {self.path} with headers={self.headers}")
        reader = pl.scan_csv if ctx.lazy else pl.read_csv
        df = reader(self.path, has_header=self.headers)
        return df



from typing import Tuple

from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class ConsoleWriter(SparkStep):
    """Prints dataframes to console"""

    async def process(self, ctx: SparkDataFrame, *dfs: SparkDataFrame) -> Tuple[SparkDataFrame]:
        ctx.logger.info("Writing to console")
        for df in dfs:
            df.show()
        return dfs



from typing import Literal

import polars as pl

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class SQLReader(PolarsStep):
    """A Generic SQL Database reader"""

    def __init__(self, sql_query: list[str] | str, connection_string: str, partition_on: str = None,
                 engine: Literal["adbc", "connectorx"] = "connectorx"):
        self.engine = engine
        self.sql_query = sql_query
        self.partition_on = partition_on
        self.connection_string = connection_string

    async def process(self, ctx: SpinePolarsContext, df: PolarsDataFrame = None) -> PolarsDataFrame:
        ctx.logger.info(f"Querying: ({self.sql_query}) using: {self.connection_string}")
        df = pl.read_database(query=self.sql_query, connection_uri=self.connection_string,
                              partition_on=self.partition_on,
                              engine=self.engine)
        return df.lazy() if ctx.lazy else df



from typing import Literal

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class SQLWriter(PolarsStep):
    """A Generic SQL Database writer
        Parameters
        ----------
        connection_string
            Connection uri, for example
            * "postgresql://username:password@server:port/database"
        if_exists : {'append', 'replace', 'fail'}
            The insert mode.
            'replace' will create a new database table, overwriting an existing one.
            'append' will append to an existing table.
            'fail' will fail if table already exists.
        engine : {'sqlalchemy', 'adbc'}
    """

    def __init__(self, table_name, connection_string: str,
                 if_exists: Literal["replace", "append", "fail"] = "replace",
                 engine: Literal["sqlalchemy", "adbc"] = "sqlalchemy"):
        self.table_name = table_name
        self.connection_string = connection_string
        self.engine = engine
        self.if_exists = if_exists

    async def process(self, ctx: SpinePolarsContext, df: PolarsDataFrame = None) -> PolarsDataFrame:
        ctx.logger.info(f"Writing to table: {self.table_name} using: {self.connection_string}]")
        df = df.collect() if ctx.lazy else df
        df.write_database(self.table_name, self.connection_string, if_exists=self.if_exists, engine=self.engine)
        return df



from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class HdfsReader(SparkStep):
    def __init__(self, path: str, format: str, schema=None, options=None):
        self.path = path
        self.format = format
        self.schema = schema
        self.options = options or {}

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame = None):
        df.spark.catalog.refreshByPath(self.path)
        data = df.spark.read.load(self.path, self.format, self.schema, **self.options)
        df.logger.info(f'Reading data from HDFS from {self.path}')
        return data



from spinelibs.spinelibs.spark.steps.io.jdbc.jdbc_writer import JDBCWriter


class MSSQLWriter(JDBCWriter):
    def __init__(self, table_name: str, connection_string: str, mode: str = "overwrite", username: str = None,
                 password: str = None, num_partitions: int = 40, batch_size: int = 50000):
        driver = "com.microsoft.sqlserver.jdbc.SQLServerDriver"
        super().__init__(table_name, mode, connection_string, username, password, driver, num_partitions, batch_size)



from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class CSVWriter(PolarsStep):
    """Local CSV file writer"""

    def __init__(self, path: str, has_headers: bool = True, sep: str = ",", batch_size: int = 1024):
        self.headers = has_headers
        self.path = path
        self.sep = sep
        self.batch_size = batch_size

    async def process(self, ctx: SpinePolarsContext, df: PolarsDataFrame = None) -> PolarsDataFrame:
        ctx.logger.info(f"Writing CSV to {self.path} with headers={self.headers}")
        writer = df.collect().write_csv if ctx.lazy else df.write_csv
        writer(self.path, has_header=self.headers, batch_size=self.batch_size, sep=self.sep)
        return df



from typing import Dict

from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class HdfsWriter(SparkStep):
    def __init__(self, path: str, mode: str, format='parquet', partitionBy: str = None, options: Dict = None):
        self.path = path
        self.format = format
        self.mode = mode
        self.partitionBy = partitionBy
        self.options = options or {}

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame):
        try:
            ctx.spark.catalog.refreshByPath(self.path)
            df.write.save(self.path, self.format, self.mode, self.partitionBy, **self.options)
            ctx.logger.info(f"Wrote dataframe to HDFS: {self.path}")
            return df
        except Exception as e:
            ctx.logger.error(f"Failed to write dataframe to HDFS: {self.path}. {type(e)}: {str(e)}")



from spinelibs.spinelibs.spark.steps.io.jdbc.jdbc_reader import JDBCReader


class MSSQLReader(JDBCReader):
    def __init__(self, query: str, connection_string: str, username: str = None, password: str = None,
                 num_partitions: int = 40, fetch_size: int = 50000):
        driver = "com.microsoft.sqlserver.jdbc.SQLServerDriver"
        super().__init__(query, connection_string, username, password, driver, num_partitions, fetch_size)



from typing import List

import polars as pl

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class GUIDColumn(PolarsStep):
    def __init__(self, column_to_hash: List[str], guid_column_name: str):
        self.guid_column_name = guid_column_name
        self.column_to_hash = column_to_hash

    async def process(self, ctx: SpinePolarsContext, df: PolarsDataFrame) -> PolarsDataFrame:
        ctx.logger.info(f"Creating GUID Coulmn ['{self.guid_column_name}'] using {self.column_to_hash}")
        return df.with_columns([pl.concat_str(*self.column_to_hash).hash().cast(pl.Utf8).alias(self.guid_column_name)])



from typing import Tuple

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.steps.polars_step import PolarsStep


class ConsoleWriter(PolarsStep):
    """Prints dataframes to console"""

    async def process(self, ctx: SpinePolarsContext, *dfs: PolarsDataFrame) -> Tuple[PolarsDataFrame]:
        ctx.logger.info("Writing to console")
        for df in dfs:
            df = df.collect() if ctx.lazy else df
            print(df)
        return dfs



from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class CsvReader(SparkStep):
    """Local CSV file reader"""

    def __init__(self, path: str, has_headers: bool):
        self.headers = has_headers
        self.path = path

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame = None) -> SparkDataFrame:
        ctx.logger.info(f"Reading {self.path} with headers={self.headers}")
        df = ctx.spark.read.csv(self.path, header=self.headers)
        return df



from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class JDBCWriter(SparkStep):
    def __init__(self, table_name: str, mode: str, connection_string: str, username: str, password: str,
                 driver: str, num_partitions: int, batch_size: int):
        self.table_name = table_name
        self.mode = mode
        self.num_partitions = num_partitions
        self.batch_size = batch_size
        self.driver = driver
        self.password = password
        self.username = username
        self.connection_string = connection_string

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame) -> SparkDataFrame:
        ctx.logger.info(f"Writing df to table: {self.table_name} with connection string: {self.connection_string}")
        df.repartition(self.num_partitions)
        df.write.format("jdbc").mode(self.mode) \
            .option("url", self.connection_string) \
            .option("numPartitions", self.num_partitions) \
            .option("batchsize", self.batch_size) \
            .option("driver", self.driver) \
            .option("dbtable", self.table_name) \
            .option("user", self.username) \
            .option("password", self.password) \
            .save()

        return df



from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class JDBCReader(SparkStep):
    def __init__(self, query: str, connection_string: str, username: str, password: str, driver: str,
                 num_partitions: int, fetch_size: int):
        self.num_partitions = num_partitions
        self.fetch_size = fetch_size
        self.driver = driver
        self.password = password
        self.username = username
        self.connection_string = connection_string
        self.query = query

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame = None) -> SparkDataFrame:
        ctx.logger.info(f"Reading df using query: {self.query} with connection string: {self.connection_string}")
        df = ctx.spark.read.format("jdbc").option("driver", self.driver) \
            .option("url", self.connection_string) \
            .option("query", self.query) \
            .option("user", self.username) \
            .option("password", self.password) \
            .option("numPartitions", self.num_partitions) \
            .option("fetchsize", self.fetch_size) \
            .load()
        return df



from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep


class SqlReader(SparkStep):
    def __init__(self, table_name: str):
        self.table_name = table_name

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame = None):
        df = ctx.spark.table(self.table_name)
        ctx.logger.info(f'Reading table: {self.table_name} from Hive MetaStore')
        return df



import pyspark.sql.functions as F
from pyspark.sql import Window

from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.steps.spark_step import SparkStep

TMP_COL = '___tmp___'
ROW_NUM_COL = '___row_num___'


class HiveWriter(SparkStep):
    def __init__(
            self, table_name: str, file_format: str = 'parquet',
            mode: str = 'overwrite', partition_by: str = None, overwrite_partition: str = False,
            num_partitions: str = None,
            is_new_partitions: str = False,
            **kwargs):
        self.table_name = table_name
        self.file_format = file_format
        self.mode = mode
        self.partition_by = partition_by
        self.kwargs = kwargs
        self.overwrite_partition = overwrite_partition
        self.num_partitions = num_partitions
        self.is_new_partitions = is_new_partitions

    async def process(self, ctx: SpineSparkContext, df: SparkDataFrame):
        try:
            if self.overwrite_partition:
                self._handle_partition_overwriting(ctx, df)
            else:
                df.write.saveAsTable(self.table_name, mode=self.mode, format=self.file_format,
                                     partitionBy=self.partition_by, **self.kwargs)
                ctx.logger.info(
                    f"Wrote table: {self.table_name} to Hive MetaStore. mode: {self.mode}, format: {self.file_format}")
        except Exception as e:
            ctx.logger.error(
                f'Failed to write table {self.table_name} to Hive. {type(e)}: {str(e)}')
            raise

        return df

    def _handle_partition_overwriting(self, ctx: SpineSparkContext, df: SparkDataFrame):
        self._validate_overwrite_partitions_params()
        ctx.spark.conf.set('spark.sql.sources.partitionOverwriteMode', 'dynamic')
        ctx = ctx.repartition(self.num_partitions)
        ctx.logger.info('set partitionOverwriteMode to dynamic')
        if self.is_new_partitions:
            self._write_one_row_per_partitions(df)
        df.select(self._get_target_table_columns(ctx)).write.insertInto(self.table_name, True)

    def _get_target_table_columns(self, cardo_context):
        return cardo_context.spark.table(self.table_name).columns

    def _validate_overwrite_partitions_params(self):
        if self.mode != 'overwrite_partitions':
            raise Exception("In order to overwrite partitions you must set mode to 'overwrite_partitions'")
        if self.num_partitions is None or self.partition_by is None:
            raise Exception("You must specify num partitions and partitions columns in partitionOverwriteMode")

    def _write_one_row_per_partitions(self, df):
        w = Window.partitionBy(self.partition_by).orderBy(TMP_COL)
        df.persist()
        tmp_df = df.withColumn(TMP_COL, F.lit(1)).withColumn(ROW_NUM_COL, F.row_number().over(w))
        tmp_df = tmp_df.where('{} = 1'.format(ROW_NUM_COL)).drop(ROW_NUM_COL).drop(TMP_COL)
        tmp_df.write.mode('append').partitionBy(self.partition_by).saveAsTable(self.table_name)
